from copy import deepcopy
from urllib.request import Request
from dateutil.parser import parse
import csv
import requests
import json
import time
from urllib3 import Retry
from urllib import parse
from toolz import first
from IPython.display import display, clear_output
import pandas as pd
import fds.analyticsapi.engines
from fds.analyticsapi.engines.api import pa_calculations_api
from fds.analyticsapi.engines.configuration import Configuration
from fds.analyticsapi.engines.api import columns_api
from fds.analyticsapi.engines.api import column_statistics_api
from fds.analyticsapi.engines.api import accounts_api
from fds.analyticsapi.engines.api import documents_api
from fds.analyticsapi.engines.api import components_api
from fds.analyticsapi.engines.api import unlinked_pa_templates_api
from fds.analyticsapi.engines.api import groups_api
from fds.analyticsapi.engines.api import frequencies_api
from fds.analyticsapi.engines.api import pricing_sources_api
from fds.analyticsapi.engines.api import currencies_api
from fds.analyticsapi.engines.api import unlinked_pa_templates_api
from fds.analyticsapi.engines.api import templated_pa_components_api
from fds.analyticsapi.engines.model.templated_pa_component_parameters_root import (
    TemplatedPAComponentParametersRoot,
)
from fds.analyticsapi.engines.model.pa_identifier import PAIdentifier
from fds.analyticsapi.engines.model.pa_calculation_column import PACalculationColumn
from fds.analyticsapi.engines.model.pa_date_parameters import PADateParameters
from fds.analyticsapi.engines.api import dates_api
from fds.analyticsapi.engines.model.pa_calculation_group import PACalculationGroup
from fds.analyticsapi.engines.model.pa_calculation_data_sources import (
    PACalculationDataSources,
)
from fds.analyticsapi.engines.model.pa_calculation_pricing_source import (
    PACalculationPricingSource,
)
from fds.analyticsapi.engines.model.pa_identifier import PAIdentifier
from fds.analyticsapi.engines.model.pa_calculation_column import PACalculationColumn
from fds.analyticsapi.engines import ApiException
from fds.analyticsapi.engines.api.pa_calculations_api import PACalculationsApi
from fds.analyticsapi.engines.model.pa_calculation_parameters_root import (
    PACalculationParametersRoot,
)
from fds.analyticsapi.engines.model.pa_calculation_parameters import (
    PACalculationParameters,
)
from fds.analyticsapi.engines.model.unlinked_pa_template_parameters import (
    UnlinkedPATemplateParameters,
)
from fds.analyticsapi.engines.model.unlinked_pa_template_parameters_root import (
    UnlinkedPATemplateParametersRoot,
)
from fds.protobuf.stach.extensions.StachVersion import StachVersion
from fds.protobuf.stach.extensions.StachExtensionFactory import StachExtensionFactory


class FDSAnalyticsAPI:
    error_assert = """Please pass your credentials (fds_username-serial, fds_apikey) to 
                    FDSAnalyticsAPI.Config.create_config before calling this function."""

    fds_api_host = "https://api.factset.com"
    http_headers = {"Content-type": "application/json", "Accept": "application/json"}

    class Config:
        @staticmethod
        def create_config(fds_username_hyphen_serial, fds_apikey):
            """
            Initialize configuraiton object for Analytics API v3
            Force 3 retrys when response times out or hits rate limit
            """
            fds_config = Configuration(
                host="https://api.factset.com",
                username=fds_username_hyphen_serial,
                password=fds_apikey,
            )
            # 429 -> Max Requests
            # 503 -> Requet Timed Out

            fds_config.retries = Retry(
                total=3,
                status=3,
                status_forcelist=frozenset([429, 503]),
                backoff_factor=2,
                raise_on_status=False,
            )
            return fds_config

    class LookUp:
        @staticmethod
        def get_absolute_date(
            fds_config, end_date, component_id, account, start_date=""
        ):
            """
            Use this function to validate the relative date ranges passed to the PA Engine.
            Requires the relative end date, componentId, and Account to verify the correct calendar is being used.
            Each specified account(.acct) or composite(.actm) must have the appropriate postfix.  Cannot be used with benchmarks or ETFs.
            Optionally, you can pass in a second (start date)
            {'enddate': '20210331', 'startdate': '20201223'}

            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                api_instance = dates_api.DatesApi(api_client)
                try:
                    # Convert PA dates to absolute format
                    api_response = api_instance.convert_pa_dates_to_absolute_format(
                        end_date, component_id, account, startdate=start_date
                    )
                    return True, first(api_response).data
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        f"Exception when calling DatesApi->convert_pa_dates_to_absolute_format: {e}",
                    )

        @staticmethod
        def get_account_dir_information(fds_config, fds_DIR_path="client:"):
            """Return a Dictionary of Accounts and Sub-directectories for the specified factset_DIR_path"""
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = accounts_api.AccountsApi(api_client)
                # example passing only required values which don't have defaults set
                try:
                    # Get accounts and sub-directories in a directory
                    api_response = api_instance.get_accounts(fds_DIR_path)
                    return True, first(api_response).data
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        f"Exception when calling AccountsApi->get_accounts: {e}",
                    )

        @staticmethod
        def get_account_schema(
            fds_username_hyphen_serial, fds_apikey, account_name_path
        ):
            """Return Account MetaData and Schema Information"""
            URL = f"{FDSAnalyticsAPI.fds_api_host}/analytics/accounts/v3/models/{FDSAnalyticsAPI.Paramaters.URL_encode(account_name_path)}/schema"
            b_success, response = FDSAnalyticsAPI.web_requests.http_get(
                URL=f"{FDSAnalyticsAPI.fds_api_host}/analytics/accounts/v3/models/{FDSAnalyticsAPI.Paramaters.URL_encode(account_name_path)}/schema",
                fds_username_hyphen_serial=fds_username_hyphen_serial,
                fds_apikey=fds_apikey,
            )
            if b_success and response.status_code == 200:
                return True, response.json(), URL
            elif response.status_code == 401 or response.status_code == 403:
                return (
                    False,
                    f"Missing or Invalid Credentials",
                    "Missing or Invalid Credentials",
                )
            elif response.status_code == 400 or response.status_code == 404:
                return (
                    False,
                    f"Please check the provided account path {account_name_path}",
                    "Provided account does not exist",
                )
            elif response.status_code == 503 or response.status_code == 404:
                return (
                    False,
                    f"Failed to retrieve Acct Information {account_name_path}",
                    "Request Timed Out",
                )
            else:
                return (
                    False,
                    f"Failed to Return Account Schema for {account_name_path}",
                    response,
                )

        @staticmethod
        def get_account_data(
            fds_username_hyphen_serial,
            fds_apikey,
            account_name_path,
            date=None,
            symbol=None,
        ):
            """
            Returns content of an Account optionally filtering the results
            by any combination of the symbol and or date
            *account_name_path is required"""
            symbol_url_part = (
                (f"symbols/{FDSAnalyticsAPI.Paramaters.URL_encode(symbol)}/")
                if symbol
                else ""
            )
            date_url_part = (
                (f"dates/{FDSAnalyticsAPI.Paramaters.format_absolute_date(date)}/")
                if date
                else ""
            )
            URL = f"{FDSAnalyticsAPI.fds_api_host}/analytics/accounts/v3/models/{FDSAnalyticsAPI.Paramaters.URL_encode(account_name_path)}/{date_url_part}{symbol_url_part}"
            b_success, response = FDSAnalyticsAPI.web_requests.http_get(
                URL=URL,
                fds_username_hyphen_serial=fds_username_hyphen_serial,
                fds_apikey=fds_apikey,
            )
            if b_success and response.status_code == 200:
                return True, response.json(), URL
            elif response.status_code == 401 or response.status_code == 403:
                return (
                    False,
                    f"Missing or Invalid Credentials",
                    response,
                )
            elif response.status_code == 400 or response.status_code == 404:
                return (
                    False,
                    f"Please check the provided account path {account_name_path}",
                    "Provided account does not exist",
                )
            else:
                return (
                    False,
                    f"Failed to Return Account Data for {account_name_path}",
                    response,
                )

        @staticmethod
        def get_calculation_parameters(fds_config, calculation_id):
            """
            Not Working
            Returns the Parameters Passed for a specified calculation id
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = pa_calculations_api.PACalculationsApi(api_client)
                # example passing only required values which don't have defaults set
                try:
                    # Get PA calculation parameters by id
                    (
                        api_response,
                        status_code,
                        *_,
                    ) = api_instance.get_calculation_parameters(calculation_id)
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return PA3 Columns",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_columns(fds_config, name="", category="", directory=""):
            """
            Return all columns that can be added when requesting
            a new calculation or creating report or templated components.
            {
                "data":{
                    "0E5D0648415D9B782078E4B048462633896171C21E919805EA94418F573E0162":{
                        "name":"Security Name",
                        "directory":"Factset",
                        "category":""
                    },
                    "05F05D9898A562CCE2F47D633C57510811552F0A911DC560ED6E3D22DECEFB95":{
                        "name":"Symbol",
                        "directory":"Factset",
                        "category":""
                    }
                }
            }
            """
            # assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = columns_api.ColumnsApi(api_client)
                try:
                    # Get PA columns
                    (api_response, status_code, *_) = api_instance.get_pa_columns(
                        name=name, category=category, directory=directory
                    )
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return PA3 Columns",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_column_statistics(fds_config):
            """
            Return All PA Column statistics
            Used when setting aggregate calculations for a pa calculatated column
            Statistics can be added to any numeric calculated column when requesting
            a new calculation or creating/modifying either report or templated components.
            {
                    "data":{
                        "eb9d6d91416e4224bacadc261787e56f":{
                            "name":"Active Weights"
                        },
                        "998e9415544f4376a7db4085fe4cda39":{
                            "name":"Average"
                        }
                    }
                }
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert
            # Enter a context with an instance of the API client
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = column_statistics_api.ColumnStatisticsApi(api_client)
                # example, this endpoint has no required or optional parameters
                try:
                    # Get PA column statistics
                    (
                        api_response,
                        status_code,
                        *_,
                    ) = api_instance.get_pa_column_statistics()
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return PA3 Column Statistics",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_components(fds_config, pa3_document):
            """
            Returns all component IDs stored within the specified pa3 document.
            For a list of PA3 documents use the get_pa3_documents() function.
            Component IDs are required for requesting a new calculation or creating/modifying report or templated components.
            {
            "data":{
                "216D0AC369E1086647CF16F33435DCCF6C9984E799A941C8CF3F2D555AC899F9":{
                    "name":"Weights",
                    "category":"Weights / Exposures"
                },
                "BA3ED6AEBB4B3885954F9A762B93691CAB7D8721663D0AA419B462C4452BDD87":{
                    "name":"Weights Difference",
                    "category":"Weights / Exposures"
                }
            }
            }
            """
            # Enter a context with an instance of the API client
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = components_api.ComponentsApi(api_client)
                # example passing only required values which don't have defaults set
                try:
                    (components, status_code, *_) = api_instance.get_pa_components(
                        pa3_document
                    )
                    return True, components.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        f"Failed To Return Component IDs",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, pa_document=pa3_document
                        ),
                    )

        @staticmethod
        def get_currency_iso(fds_config):
            """
            Return all avaialable Currency ISO Names, and their codes.
            Currency ISO codes can be used to overwrite the default currency when
            requesting a new calculation or creating new report or templated component.
            {
                "data":{
                    "AFN":{
                        "name":"Afghanistan Afghani"
                    },
                    "ALL":{
                        "name":"Albanian Lek"
                    },
                    "DZD":{
                        "name":"Algerian Dinar"
                    }
                }
            }
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = currencies_api.CurrenciesApi(api_client)
                try:
                    # Get PA columns
                    (api_response, status_code, *_) = api_instance.get_currencies()
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return PA3 Currency ISOs",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_default_unlinked_PA_template_types(fds_config):
            """
            Returns of a list of default unlinked PA Template Types
            All Unlinked PA Template must be linked to a default PA Template Type.
            {
                "data":{
                    "996E90B981AEE83F14029ED3D309FB3F03EC6E2ACC7FD42C22CBD5D279502CFD":{
                        "category":"Weights",
                        "name":"Weights"
                    },
                    "A4BD67A4906BC0B02F3F72F25E6B7D85838F9B69C02BA0147F811747BB05B151":{
                        "category":"Weights",
                        "name":"Weights Difference"
                    }
                }
            }
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                    api_client
                )

                # example, this endpoint has no required or optional parameters
                try:
                    # Get default unlinked PA template types.
                    (
                        api_response,
                        status_code,
                        *_,
                    ) = api_instance.get_default_unlinked_pa_template_types()
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        f"Failed to Return Unlinked Template Types",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_frequency_ids(fds_config):
            """
            Returns a full list of available frequency identifiers that can be applied to a calculation, component, or report template.
            {
                "data":{
                    "Single":{
                        "name":"Single"
                    },
                    "Daily":{
                        "name":"Daily"
                    }
                }
            }
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = frequencies_api.FrequenciesApi(api_client)
                try:
                    # Get PA frequencies
                    (api_response, status_code, *_) = api_instance.get_pa_frequencies()
                    return True, api_response.data, ""

                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return PA3 Frequencies",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_groups(fds_config):
            """
            Retuns all Available groupings
            list of objects used to create a grouping hierarchy when requesting a new calculation or creating report or templated component.
            The first listed grouping id will be the outer most grouping level.
            {
                "data":{
                    "10D46537D08F37A521A57FC77105FFE7CA8F336AAFDF876B09EAA384A6024ED0":{
                        "name":"3M Momentum",
                        "directory":"Client",
                        "category":"Client"
                    },
                    "07D67EFA4356CEA4CFE65F545BB6B158BB6A8356D7C5F89A7544CDAC4A0936F2":{
                        "name":"AGI Asset Class",
                        "directory":"Client",
                        "category":"Client"
                    }
                }
            }
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = groups_api.GroupsApi(api_client)
                # example, this endpoint has no required or optional parameters
                try:
                    # Get PA groups
                    (api_response, status_code, *_) = api_instance.get_pa_groups()
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return PA3 Groupings",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_PA3_documents(fds_config, path=""):
            """
            Returns all sub-directories, Accounts (acct) and Composites (actm) for the specified factset directory (path).
            The path will start with “factset:”, “client:”, “personal:”, or “super_client:”
            Results can be used as an input to the get_components() function.
            {
                "data":{
                    "documents":[
                        "Academic Default Reports",
                        "Decision Analysis",
                        "Default"
                    ]
                }
            }
            """
            assert bool(fds_config), FDSAnalyticsAPI.error_assert
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = documents_api.DocumentsApi(api_client)
                # example passing only required values which don't have defaults set
                try:
                    # Get PA3 documents and sub-directories in a directory
                    (api_response, status_code, *_) = api_instance.get_pa3_documents(
                        path
                    )
                    return True, api_response.data, ""

                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed To Return PA3 Documents",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_pricing_sources(fds_config, directory="", category="", name=""):
            """
            Returns a full list of all the PA pricing sources that can be applied to a PA calculation

            "data": {
                "28D31D9E69D4B8A4A5E7370AEC73B3A124B7BDB4B6F8ED74962F161E551E186F": {
                    "name": "Super Client Security Master",
                    "directory": "Client Provided",
                    "category": ""
                }
             }
            """
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = pricing_sources_api.PricingSourcesApi(api_client)
                try:
                    # Get PA columns
                    (
                        api_response,
                        status_code,
                        *_,
                    ) = api_instance.get_pa_pricing_sources(
                        name=name, category=category, directory=directory
                    )
                    return True, api_response.data, ""
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Failed to Return Pricing Sources",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

        @staticmethod
        def get_unlinked_pa_templates(fds_config, category=""):
            """
            Returns a full list of unlinked PA templates stored in the specified FactSet directory.
            The directory path will start with “client:”, “personal:”, or “super_client:”.
            This function optionally allows to filter the unlinked pa template by either the directory or category.
            *Currenlty unlinked pa templates can only be saved in the personal directory
            {
                "630156ba0bef442c9f80a4eddf990a03":{
                    "category":"Weights",
                    "description":"Overlap Report",
                    "name":"Personal:630156ba0bef442c9f80a4eddf990a03.UNLINKED_PA_TEMPLATE"
                }
            }
            """
            directory = ""
            assert bool(fds_config), FDSAnalyticsAPI.error_assert
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                    api_client
                )
                # example passing only required values which don't have defaults set
                # and optional values
                if category == "":
                    directory = "personal:"
                try:
                    # Get unlinked PA templates
                    (
                        api_response,
                        status_code,
                        *_,
                    ) = api_instance.get_unlinked_pa_templates(
                        directory=directory, category=category
                    )
                    return True, api_response.data, ""

                except fds.analyticsapi.engines.ApiException as e:

                    return (
                        False,
                        "Failed to Return Unlinked PA Templates",
                        FDSAnalyticsAPI.Requests.get_err_detail(
                            status_code=e.status, is_lookup=True
                        ),
                    )

    class Paramaters:
        @staticmethod
        def make_PA_calculation_column(id, statistics):
            return PACalculationColumn(id=id, statistics=statistics)

        @staticmethod
        def make_PA_calculation_group(id):
            return PACalculationGroup(id=id)

        # @staticmethod
        def make_PA_data_source(
            portfolio_pricing_sources,
            benchmark_pricing_sources,
            use_portfolio_pricing_source=False,
        ):
            #     portfolio_pricing_source =
            if portfolio_pricing_sources and use_portfolio_pricing_source:
                return True, PACalculationDataSources(
                    portfoliopricingsources=portfolio_pricing_sources,
                    useportfoliopricingsourcesforbenchmark=True,
                )
            elif use_portfolio_pricing_source:
                return True, PACalculationDataSources(
                    useportfoliopricingsourcesforbenchmark=True
                )
            elif portfolio_pricing_sources and benchmark_pricing_sources:
                return True, PACalculationDataSources(
                    portfoliopricingsources=portfolio_pricing_sources,
                    benchmarkpricingsources=benchmark_pricing_sources,
                )
            elif portfolio_pricing_sources:
                return True, PACalculationDataSources(
                    portfoliopricingsources=portfolio_pricing_sources
                )
            elif benchmark_pricing_sources:
                return True, PACalculationDataSources(
                    benchmarkpricingsources=benchmark_pricing_sources
                )

            return False, "No Sources Specified"

        def make_PA_princing_source(source_id):
            return PACalculationPricingSource(id=source_id)

        @staticmethod
        def make_PA_calculations_paramaters(component_id):
            """Returns PACalculationParameter Object"""
            return PACalculationParameters(componentid=component_id)

        @staticmethod
        def make_PA_calculation_parameters_root(calculation_parameters):
            return PACalculationParametersRoot(data=calculation_parameters)

        @staticmethod
        def make_PA_date_parameters(end_date, frequency, start_date=""):
            if start_date:
                return PADateParameters(
                    startdate=FDSAnalyticsAPI.Paramaters.__check_date_format(
                        start_date
                    ),
                    enddate=FDSAnalyticsAPI.Paramaters.__check_date_format(end_date),
                    frequency=frequency,
                )
            else:
                return PADateParameters(
                    enddate=FDSAnalyticsAPI.Paramaters.__check_date_format(end_date),
                    frequency=frequency,
                )

        @staticmethod
        def make_PA_identifier(PA_Identifier):
            return PAIdentifier(id=PA_Identifier)

        @staticmethod
        def make_unlinked_PA_template_parameters(
            directory, template_type_id, description
        ):
            return UnlinkedPATemplateParameters(
                directory=directory,
                template_type_id=template_type_id,
                description=description,
            )

        @staticmethod
        def make_unlinked_PA_template_parameters_root(data, meta):
            return UnlinkedPATemplateParametersRoot(data=data, meta=meta)

        @staticmethod
        def set_attribute(parameters, name, object):
            if object:
                parameters.set_attribute(name, object)

        # Additional Functions Used for Creating Parameters
        @staticmethod
        def __check_date_format(date_value):
            """I have to account for "M" in the input.  The parser library treats -1AM as a valid date
            This function trys to determine if a date is in relative or absolute date format and then format the absolute dates to YYYYMMDD
            """
            try:
                if not "M" in str(date_value).upper():
                    return FDSAnalyticsAPI.Paramaters.format_absolute_date(date_value)
                else:
                    return date_value
            except:
                return date_value

        @staticmethod
        def URL_encode(value_to_encode):
            """Encodes String Replace Special Characters for URL Encoding"""
            return parse.quote(value_to_encode, safe="")

        @staticmethod
        def format_absolute_date(date_value):
            "Format date to YYYYMMDD format"
            try:
                return parse(date_value).strftime("%Y%m%d")
            except:
                return date_value

    class Requests:
        @staticmethod
        def make_modify_model_portfolio(
            fds_username_hyphen_serial, fds_apikey, account_name_path, json_string, URL
        ):
            """Create a New or Update and Existing Account"""

            bSuccess, response = FDSAnalyticsAPI.web_requests.http_put(
                URL=URL,
                json_string=json_string,
                fds_username_hyphen_serial=fds_username_hyphen_serial,
                fds_apikey=fds_apikey,
            )
            if bSuccess and response.status_code in (201, 200):
                return (
                    True,
                    f"Model Portfolio {account_name_path} Successfully {'Created' if response.status_code==201 else 'Updated'} (Status Code: {response.status_code})",
                    "",
                )
            else:
                errMsg = (
                    f"Failed to Create Account (Status Code: {response.status_code})"
                )
                errDetail = FDSAnalyticsAPI.Requests.get_err_detail(
                    response.status_code
                )
                return False, errMsg, errDetail

        @staticmethod
        def get_err_detail(
            status_code, is_docbase=False, is_lookup=False, pa_document=""
        ):
            if status_code == 204:
                return "Model Account does not exist"
            elif status_code == 400 and is_docbase:
                return "Invalid PA document name or Query Parameter"
            elif status_code == 400 and is_lookup:
                return "Invalid Input Parameter(s)"
            elif status_code == 400:
                return "Invalid POST Body or Account Information"
            elif status_code == 401:
                return "Missing or Invalid Authentication"
            elif status_code == 403:
                return "User forbidden with Current Credentials"
            elif status_code == 404:
                return (
                    f"Docuement Not Found: {pa_document}"
                    if (pa_document)
                    else "Document Not Found"
                )
            elif status_code == 500:
                return "Request timed out."
            else:
                return ""

        @staticmethod
        def make_model_portfolio_json_from_CSV(CSV_file_path, delimiter=","):

            """
            Expects Text file to be CSV and First Row are the dictionary keys.
            date,symbol,weight,price,priceiso
            20220203,FDS,25,1,usd
            20220203,AAPL,25,2,usd
            20220203,TSLA,25,3,usd
            20220203,IBM,25,4,usd

            Must have a date and symbol column

            """
            csv_model_portfolio_json = {}
            data = {}
            symbol_index = None
            date_index = None
            counter = 0
            column_list = []
            with open(CSV_file_path) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=delimiter)
                bFirst = True
                for row in csv_reader:
                    if not row:
                        continue
                    else:
                        if bFirst == True:
                            for column in row:
                                if column.upper().strip() == "SYMBOL":
                                    symbol_index = counter
                                elif column.upper().strip() == "DATE":
                                    date_index = counter
                                column_list.append(column)
                                counter += 1
                            bFirst = False
                        else:
                            if not symbol_index is None and not date_index is None:
                                csv_model_portfolio_json = (
                                    FDSAnalyticsAPI.Requests.__check_mp_key(
                                        dict_data=csv_model_portfolio_json,
                                        key=row[date_index],
                                    )
                                )
                                csv_model_portfolio_json[row[date_index]][
                                    row[symbol_index]
                                ] = FDSAnalyticsAPI.Requests.__create_mp_object(
                                    column_list, row, symbol_index, date_index
                                )
                            else:
                                return False, "Missing Date or Symbol Column"

            return True, {"data": {"iterative": csv_model_portfolio_json}}

        @staticmethod
        def __check_mp_key(dict_data, key):
            if key not in dict_data:
                dict_data[key] = {}
            return dict_data

        @staticmethod
        def __create_mp_object(list_keys, list_values, symbol_index, date_index):
            dict_results = {}
            for x in range(len(list_keys)):
                if x != symbol_index and x != date_index:
                    dict_results[list_keys[x]] = list_values[x]
            return dict_results

        @staticmethod
        def make_PA_engine_request(fds_config, PA_calculation_parameter_root):
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                pa_calculations_api = PACalculationsApi(api_client)
                try:
                    pa_calculation_parameter_root = PA_calculation_parameter_root
                    pa_calculations_api = PACalculationsApi(api_client)
                    post_and_calculate_response = (
                        pa_calculations_api.post_and_calculate(
                            pa_calculation_parameters_root=pa_calculation_parameter_root
                        )
                    )
                    # Check if Long Running
                    return True, post_and_calculate_response, pa_calculations_api
                except ApiException as e:
                    return False, f"Failed {e}", pa_calculations_api

        @staticmethod
        def wait_for_long_running_response(calculation_id, pa_calculations_api):
            try:
                # print("Calculation Id: " + calculation_id)
                status_response = pa_calculations_api.get_calculation_status_by_id(
                    id=calculation_id
                )
                # Loop while waiting for long running request
                while status_response[1] == 202 and (
                    first(status_response).data.status in ("Queued", "Executing")
                ):
                    max_age = "5"
                    age_value = status_response[2].get("cache-control")
                    if age_value is not None:
                        max_age = age_value.replace("max-age=", "")
                    # print("Sleeping: " + max_age)
                    print(f"Sleeping for {max_age} seconds.")
                    time.sleep(int(max_age))
                    status_response = pa_calculations_api.get_calculation_status_by_id(
                        calculation_id
                    )
                return True, status_response
            except ApiException as e:
                return False, "Api exception Encountered", ""

        @staticmethod
        def return_long_running_response(
            status_response, calculation_id, pa_calculations_api
        ):
            """
            Returns the first completed table from a long running request
            """
            # Check Calc Units to See if Completed
            list_results = []
            try:
                for (
                    calculation_unit_id,
                    calculation_unit,
                ) in status_response[0].data.units.items():
                    if calculation_unit.status == "Success":
                        result_response = (
                            pa_calculations_api.get_calculation_unit_result_by_id(
                                id=calculation_id, unit_id=calculation_unit_id
                            )
                        )
                        list_results.append(
                            {
                                "isSuccess": True,
                                "results": first(result_response)["data"],
                                "calc_unit_id": calculation_unit_id,
                            }
                        )
                    else:
                        list_results.append(
                            {
                                "isSuccess": False,
                                "results": f"""Calculation Unit Id: {calculation_unit_id} Failed!!! Error message : {str(calculation_unit.errors)}""",
                                "calc_unit_id": calculation_unit_id,
                            }
                        )
                return list_results

            except ApiException as e:
                return False, "Api exception Encountered"

        @staticmethod
        def make_templated_component_request(fds_config, data):
            """Create new templated component"""
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = templated_pa_components_api.TemplatedPAComponentsApi(
                    api_client
                )
                templated_pa_component_parameters_root = (
                    TemplatedPAComponentParametersRoot(
                        data=data,
                        meta=None,
                    )
                )
                try:
                    # Create templated PA component
                    api_response = api_instance.create_templated_pa_components(
                        templated_pa_component_parameters_root
                    )
                    return True, first(api_response).data
                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        f"Exception when calling TemplatedPAComponentsApi->create_templated_pa_components:{e}",
                    )

        @staticmethod
        def make_unlinked_pa_template_request(fds_config, data):
            """Create new unlinked pa template"""
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                    api_client
                )
                unlinked_pa_template_parameters_root = UnlinkedPATemplateParametersRoot(
                    data=data,
                    meta=None,
                )
                try:
                    # Create unlinked PA template
                    (
                        api_response,
                        response_status,
                        *_,
                    ) = api_instance.create_unlinked_pa_templates(
                        unlinked_pa_template_parameters_root
                    )
                    if response_status == 201:
                        dict_response = api_response.to_dict()
                        return (
                            True,
                            f"Unlinked PA Template Successfully Created - {dict_response['data']['id']}  (Stauts Code: 201)",
                            dict_response,
                        )
                    else:
                        return (
                            False,
                            "Failed to Create Unlinked PA Template",
                            FDSAnalyticsAPI.Requests.get_err_detail(api_response[1]),
                            dict_response,
                        )

                except fds.analyticsapi.engines.ApiException as e:
                    return (
                        False,
                        "Exception when calling UnlinkedPATemplatesApi",
                        e,
                    )

        @staticmethod
        def delete_account(fds_username_hyphen_serial, fds_apikey, account_name_path):
            """Delete and Existing Account"""
            URL = f"{FDSAnalyticsAPI.fds_api_host}/analytics/accounts/v3/models/{FDSAnalyticsAPI.Paramaters.URL_encode(account_name_path)}"

            bSuccess, response = FDSAnalyticsAPI.web_requests.http_delete(
                URL=URL,
                fds_username_hyphen_serial=fds_username_hyphen_serial,
                fds_apikey=fds_apikey,
            )
            # Worked
            if bSuccess and response.status_code == 204:
                return (
                    True,
                    f"Model Portfolio {account_name_path} Successfully Deleted (Status Code: 204)",
                    URL,
                )
            # Failed without error
            elif bSuccess:
                errMsg = f"Failed to Delete Account {account_name_path} (Status Code: {response.status_code})"
                errDisc = FDSAnalyticsAPI.Requests.get_err_detail(
                    response.status_code
                )
                return (
                    False,
                    errMsg,
                    errDisc,
                )
            # failed
            else:
                return False, f"Failed to Delete Account {account_name_path}", ""

        @staticmethod
        def delete_calculation_by_id(fds_config, id):
            """Delete A calculation request before completetion"""
            assert bool(fds_config), FDSAnalyticsAPI.error_assert

            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = pa_calculations_api.PACalculationsApi(api_client)
                try:
                    # Cancel PA calculation by id
                    api_instance.cancel_calculation_by_id(id)
                    return True, f"Calculation {id} Successfully Deleted"
                except fds.analyticsapi.engines.ApiException as e:
                    raise ValueError(False, f"Failed to dete Calculation {id} {e}")

        @staticmethod
        def delete_unlinked_pa_template(fds_config, id):
            """Delete Unlinked PA Report Template"""
            # Enter a context with an instance of the API client
            with fds.analyticsapi.engines.ApiClient(fds_config) as api_client:
                # Create an instance of the API class
                api_instance = unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                    api_client
                )
                try:
                    # delete unlinked PA template
                    api_response = api_instance.delete_unlinked_pa_templates(id=id)
                    return True, "Unlinked Template Deleted Successfully", ""
                except fds.analyticsapi.engines.ApiException as e:
                    #print(json.dumps(first(json.loads(e.body)["errors"])["detail"],indent=2))
                    return False, "Failed to Delete Unlinked Template", first(json.loads(e.body)["errors"])["detail"]

    class web_requests:
        @staticmethod
        def http_delete(URL, fds_username_hyphen_serial, fds_apikey):
            try:
                response = requests.delete(
                    url=URL,
                    auth=(fds_username_hyphen_serial, fds_apikey),
                    headers=FDSAnalyticsAPI.http_headers,
                )

                return True, response

            except Exception as e:
                return False, e

        @staticmethod
        def http_get(URL, fds_username_hyphen_serial, fds_apikey):
            try:
                return True, requests.get(
                    URL,
                    auth=(fds_username_hyphen_serial, fds_apikey),
                    headers=FDSAnalyticsAPI.http_headers,
                )
            except Exception as e:
                return False, (e)

        # post web request
        @staticmethod
        def http_put(URL, json_string, fds_username_hyphen_serial, fds_apikey):
            try:
                return True, requests.put(
                    URL,
                    auth=(fds_username_hyphen_serial, fds_apikey),
                    headers=FDSAnalyticsAPI.http_headers,
                    data=json.dumps(json_string),
                )
            except Exception as e:
                return False, str(e)

    class FormatResults:
        @staticmethod
        def format_dataframe(df):

            """Reindex and Center Headers"""
            header_align = "center"
            text_align = "left"
            return (df.style.hide(axis="index")).set_table_styles(
                [
                    dict(selector="td", props=[("text-align", text_align)]),
                    dict(selector="th", props=[("text-align", header_align)]),
                ]
            )

        @staticmethod
        def return_calculation_results_as_dataframe(result):
            """Convert Column Based Stach into DataFrame (List)"""
            stachBuilder = StachExtensionFactory.get_row_organized_builder(
                StachVersion.V2
            )
            stachExtension = stachBuilder.set_package(result).build()
            dataFramesList = stachExtension.convert_to_dataframe()
            return dataFramesList

        @staticmethod
        def return_directory_category_dictionary(
            pa_dictionary, is_tuple=False, key_value="", display_value="name"
        ):
            results = {}
            results[""] = {"": [("", "")]}
            for k, v in pa_dictionary.items():
                category = v["category"] if not v["category"] == "" else "uncategorized"
                directory = (
                    v["directory"] if not v["directory"] == "" else "No Directory"
                )
                if key_value:
                    component_display = (
                        f"{v[display_value]}",
                        (
                            f"{v[display_value]} ({k})",
                            (directory, category, v[display_value], k, key_value)
                            if not is_tuple
                            else (directory, category, v[display_value], k, []),
                        ),
                    )
                else:
                    component_display = (
                        f"{v[display_value]}",
                        (
                            f"{v[display_value]} ({k})",
                            (directory, category, v[display_value], k)
                            if not is_tuple
                            else (directory, category, v[display_value], k, []),
                        ),
                    )
                if directory in results:
                    if category in results[directory]:
                        results[directory][category].append(component_display)
                    else:
                        results[directory][category] = [("", "")] + [component_display]
                else:
                    results[directory] = {"": [("", "")]}
                    results[directory][category] = [("", "")] + [component_display]

            return dict(sorted(results.items()))

        @staticmethod
        def return_categorized_dictionary(
            pa_dictionary, is_tuple=False, key_value="", display_value="name"
        ):
            """Return Dictionary of Tuples Used to Populate the Category Item Box Widget Combination Below"""
            results = {}
            results[""] = [("", "")]
            for k, v in pa_dictionary.items():
                category = v["category"] if not v["category"] == "" else "uncategorized"
                if key_value:
                    component_display = (
                        f"{v[display_value]}",
                        (
                            f"{v[display_value]} ({k})",
                            (category, v[display_value], k, key_value)
                            if not is_tuple
                            else (category, v[display_value], k, []),
                        ),
                    )
                else:
                    component_display = (
                        f"{v[display_value]}",
                        (
                            f"{v[display_value]} ({k})",
                            (category, v[display_value], k)
                            if not is_tuple
                            else (category, v[display_value], k, []),
                        ),
                    )
                if category in results:
                    results[category].append(component_display)
                else:
                    results[category] = [("", "")] + [component_display]
            return dict(sorted(results.items()))

        @staticmethod
        def return_dataframe_from_model_portfolio_response(json):
            output = []

            for row in json["data"]["data"]["rows"]:
                output.append(row["values"])

            return pd.DataFrame.from_dict(output)

        @staticmethod
        def return_lookup_results_as_dataframe(results):
            """format Analytics API LookUp Endpoints as Dataframes"""
            df = pd.DataFrame(
                [dict(v.to_dict(), **{"id": k}) for k, v in results.items()]
            ).rename(str.upper, axis="columns")
            return df

        @staticmethod
        def return_uncategorized_dictionary(pa_dictionary):
            """Return TUPLE for uncategorized reference data from FactSet
            Append Empty Value in First Position"""
            results = []
            results.append(("", ""))

            for k, v in pa_dictionary.items():
                results.append((f"{v['name']} ({k})", k))

            return results

        @staticmethod
        def truncate_and_display_tables(df, max_rows_returned=20):
            """Format Resulting Tables and Truncate Results to make viewing easier
            Returns if Truncated Dataframe and the Rows Skipped"""
            if df.shape[0] <= max_rows_returned or max_rows_returned < 0:
                return False, FDSAnalyticsAPI.FormatResults.format_dataframe(df), ""
            else:
                return (
                    True,
                    FDSAnalyticsAPI.FormatResults.format_dataframe(
                        df.head(max_rows_returned)
                    ),
                    f"{str(len(df.index) - max_rows_returned)} Rows skipped!",
                )
