import ipywidgets as widgets
import json as JSON
from toolz import first
from IPython.display import display, clear_output
from Analytics_API_Common_Functions import FDSAnalyticsAPI
from Common_Widgets import (
    BasicWidgets,
    WidgetContainers,
    WidgetLayouts,
    WidgetCombinations,
)

class AnalyticsAPIWidgets:
    fds_config = None
    max_rows_returned = 20
    output = None
    __fds_username_hyphen_serial = None
    __fds_apikey = None

    def __init__(self, fds_username_hyphen_serial, fds_apikey):
        self.fds_config = FDSAnalyticsAPI.Config.create_config(
            fds_username_hyphen_serial=fds_username_hyphen_serial, fds_apikey=fds_apikey
        )
        self.__fds_apikey = fds_apikey
        self.__fds_username_hyphen_serial = fds_username_hyphen_serial
        self.output = widgets.Output()

    # Common Visuals
    def make_columns_box(self, default_value):

        description_label = BasicWidgets.make_description_label(
            self.AnalyticsAPIDefaultValues.columns_label
        )
        b_columns, response, response_err = FDSAnalyticsAPI.LookUp.get_columns(
            fds_config=self.fds_config
        )
        (
            b_stats,
            statistics_dictionary,
            statistics_err,
        ) = FDSAnalyticsAPI.LookUp.get_column_statistics(fds_config=self.fds_config)

        if b_columns and b_stats:
            columns = FDSAnalyticsAPI.FormatResults.return_categorized_dictionary(
                pa_dictionary=response, is_tuple=True
            )
            statistics = FDSAnalyticsAPI.FormatResults.return_uncategorized_dictionary(
                statistics_dictionary
            )
            (
                category_box,
                columns_selectmultiple,
            ) = WidgetCombinations.make_category_item_box(
                components_dictionary=columns,
                category_label="Column Category",
                item_label="Columns",
                selected_label="Selected Columns",
                uncategorized_additional_dictionary=statistics,
                additional_info_label_text="Select One or More Statistics",
            )
            columns_selectmultiple.options = default_value

            return columns_selectmultiple, WidgetContainers.make_vbox(
                children=[description_label, category_box],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
        elif b_stats:
            return BasicWidgets.make_error_label(response, response_err)
        else:
            return BasicWidgets.make_error_label(statistics_dictionary, statistics_err)

    def make_component_detail_box(self, default_value):
        descroption_label = BasicWidgets.make_description_label(
            self.AnalyticsAPIDefaultValues.component_detail_label
        )
        list_details = [
            ("", ""),
            ("Securities", "securities"),
            ("Groups", "groups"),
            ("Totals", "totals"),
        ]
        (
            dropdown_box,
            component_detail_dropdwn,
        ) = WidgetCombinations.make_dropdown_with_label(
            label="Component Detail",
            options=list_details,
            value=default_value,
        )

        return component_detail_dropdwn, WidgetContainers.make_vbox(
            children=[descroption_label, dropdown_box],
            layout=WidgetLayouts.make_tab_box_layout(),
        )

    def make_currency_iso_box(self, default_value):

        description_label = BasicWidgets.make_description_label(
            self.AnalyticsAPIDefaultValues.currency_label
        )
        b_currency, response, response_err = FDSAnalyticsAPI.LookUp.get_currency_iso(
            fds_config=self.fds_config
        )
        if b_currency:

            currency = FDSAnalyticsAPI.FormatResults.return_uncategorized_dictionary(
                pa_dictionary=response
            )
            (
                currency_box,
                currecny_iso_dropdown,
            ) = WidgetCombinations.make_dropdown_with_label(
                "Currency ISO",
                options=currency,
                value=default_value,
            )
            return currecny_iso_dropdown, WidgetContainers.make_vbox(
                children=[description_label, currency_box],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
        else:
            return BasicWidgets.make_error_label(response, response_err)

    def make_dates_box(self, start_date, end_date, frequency):

        (
            b_frequencies,
            response,
            response_err,
        ) = FDSAnalyticsAPI.LookUp.get_frequency_ids(fds_config=self.fds_config)

        if b_frequencies:
            frequencies = FDSAnalyticsAPI.FormatResults.return_uncategorized_dictionary(
                pa_dictionary=response
            )

            description_label = BasicWidgets.make_description_label(
                self.AnalyticsAPIDefaultValues.date_description
            )
            (
                start_date_box,
                start_date_textbox,
            ) = WidgetCombinations.make_textbox_with_label(
                label="Start Date (Oldest)",
                value=start_date,
                layout=WidgetLayouts.make_layout(
                    margin="0px 10px 0px 0px", width="120px"
                ),
                text_box_width="120px",
            )
            end_date_box, end_date_textbox = WidgetCombinations.make_textbox_with_label(
                label="End Date (Earliest)",
                value=end_date,
                layout=WidgetLayouts.make_layout(
                    margin="0px 10px 0px 0px", width="120px"
                ),
                text_box_width="120px",
            )
            freq_box, freq_dropdown = WidgetCombinations.make_dropdown_with_label(
                label="Frequency",
                options=frequencies,
                value=frequency,
                layout=WidgetLayouts.make_layout(min_width="250px", height="28px"),
            )

            date_box = WidgetContainers.make_hbox(
                children=[start_date_box, end_date_box, freq_box]
            )
            wrapper_box = WidgetContainers.make_vbox(
                children=[description_label, date_box],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
            date_object = {
                "start_date_textbox": start_date_textbox,
                "end_date_textbox": end_date_textbox,
                "freq_dropdown": freq_dropdown,
            }

            return date_object, wrapper_box
        else:
            return BasicWidgets.make_error_label(response, response_err)

    def make_groups_box(self, default_value):
        b_groups, response, response_err = FDSAnalyticsAPI.LookUp.get_groups(
            fds_config=self.fds_config
        )

        if b_groups:
            groups = FDSAnalyticsAPI.FormatResults.return_directory_category_dictionary(
                response
            )
            description_label = BasicWidgets.make_description_label(
                self.AnalyticsAPIDefaultValues.group_label
            )
            (
                group_box,
                group_box_selectmultiple,
            ) = WidgetCombinations.make_directory_category_item_box(
                description="",
                components_dictionary=groups,
                directory_label="Directory",
                category_label="Group Categories",
                item_label="Groups",
                single_select=False,
            )


            # (
            #     template_type_box,
            #     group_box_selectmultiple,
            # ) = WidgetCombinations.make_category_item_box(
            #     components_dictionary=groups,
            #     category_label="Group Categories",
            #     item_label="Groups",
            #     selected_label="Selected Groups",
            #     single_select=False,
            # )
            group_box_selectmultiple.options = default_value
            return group_box_selectmultiple, WidgetContainers.make_vbox(
                children=[description_label, group_box],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
        else:
            return BasicWidgets.make_error_label(response, response_err)

    def make_portfolio_benchmark_box(self, default_portfolio, default_benchmark):
        portfolio_benchmark_label = BasicWidgets.make_description_label(
            self.AnalyticsAPIDefaultValues.portfolio_benchmark_label
        )
        (
            portfolio_box,
            portfolio_selectmultiple,
        ) = WidgetCombinations.make_text_input_box(
            text_label="Full Portfolio Path",
            text_value=default_portfolio,
            selectmultiple_label="Selected Accout(s)",
            layout=WidgetLayouts.make_layout(margin="0px 10px 0px 0px"),
        )
        (
            benchmark_box,
            benchmark_selectmultiple,
        ) = WidgetCombinations.make_text_input_box(
            text_label="Full Benchmark Path",
            text_value=default_benchmark,
            selectmultiple_label="Selected Benchmark(s)",
        )
        portfolio_selectmultiple.options = [
            (
                default_portfolio,
                default_portfolio,
            )
        ]
        benchmark_selectmultiple.options = [
            (
                default_benchmark,
                default_benchmark,
            )
        ]
        portfolio_benchmark_object = {
            "portfolio_selectmultiple": portfolio_selectmultiple,
            "benchmark_selectmultiple": benchmark_selectmultiple,
        }
        return portfolio_benchmark_object, WidgetContainers.make_vbox(
            children=[
                portfolio_benchmark_label,
                WidgetContainers.make_hbox(
                    children=[portfolio_box, benchmark_box],
                ),
            ],
            layout=WidgetLayouts.make_tab_box_layout(),
        )

    def make_pricing_sources_box(self, default_value=""):
        b_princing, response, response_err = FDSAnalyticsAPI.LookUp.get_pricing_sources(
            fds_config=self.fds_config
        )

        if b_princing:
            pricing_sources = (
                FDSAnalyticsAPI.FormatResults.return_directory_category_dictionary(
                    response
                )
            )
            description_label = BasicWidgets.make_description_label(
                self.AnalyticsAPIDefaultValues.pricing_label
            )
            (
                portfolio_box,
                portfolio_selectmultiple,
            ) = WidgetCombinations.make_directory_category_item_box(
                description="Portfolio Pricing Sources",
                components_dictionary=pricing_sources,
                directory_label="Directory",
                category_label="Category",
                item_label="Pricing Source",
                single_select=False,
            )
            (
                benchmark_box,
                benchmark_selectmultiple,
            ) = WidgetCombinations.make_directory_category_item_box(
                description="Benchmark Pricing Sources",
                components_dictionary=pricing_sources,
                directory_label="Directory",
                category_label="Category",
                item_label="Pricing Source",
                single_select=False,
            )
            if default_value:
                portfolio_selectmultiple.options = default_value
                benchmark_selectmultiple.options = default_value

            check_box = BasicWidgets.make_checkbox(
                description="Apply Same Pricing Sources to Both"
            )

            def change(change):
                if change["type"] == "change" and change["name"] == "value":
                    benchmark_box.layout.visibility = (
                        "hidden" if change["new"] == True else "visible"
                    )

            check_box.observe(change)

            pricing_object = {
                "portfolio_selectmultiple": portfolio_selectmultiple,
                "benchmark_selectmultiple": benchmark_selectmultiple,
                "check_box": check_box,
            }
            return pricing_object, WidgetContainers.make_vbox(
                children=[
                    description_label,
                    WidgetContainers.make_hbox(
                        children=[portfolio_box, benchmark_box],
                        layout=WidgetLayouts.make_layout(width="max-content"),
                    ),
                    check_box,
                ],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
        else:
            return BasicWidgets.make_error_label(response, response_err)

    # Common Outputs
    def make_pa_calculation_request(
        self,
        calculation_parameter_root,
        unlinked_component_id="",
        return_results_as_json=False,
    ):
        """'Used to Keep User Informed during Long running requests"""
        results = []
        with self.output:
            (
                bRequest,
                PA_results,
                pa_calculations_api,
            ) = FDSAnalyticsAPI.Requests.make_PA_engine_request(
                fds_config=self.fds_config,
                PA_calculation_parameter_root=calculation_parameter_root,
            )
            # Check if Valid
            if bRequest:
                if PA_results[1] != 201:
                    display(
                        BasicWidgets.make_success_request_label(
                            text=f"PA Calculation Request Successfully Created - {first(PA_results).data.calculationid} (Status Code: {PA_results[1]})"
                        )
                    )
                if PA_results[1] == 200:
                    # Assume all Multi-Requests will always be long Running when returning 200
                    # If 200 and single it failed
                    if len(calculation_parameter_root["data"].keys()) > 1:
                        calculation_id = first(PA_results).data.calculationid
                        results = FDSAnalyticsAPI.Requests.return_long_running_response(
                            status_response=PA_results,
                            calculation_id=calculation_id,
                            pa_calculations_api=pa_calculations_api,
                        )

                    else:
                        for (
                            calculation_unit_id,
                            calculation_unit,
                        ) in first(PA_results).data.units.items():
                            results.append(
                                {
                                    "isSuccess": False,
                                    "results": f"Calculation Unit Id: {calculation_unit_id} Failed!!!: {(calculation_unit.errors)}",
                                    "calc_unit_id": calculation_unit_id,
                                }
                            )

                # If short running return results from body Assume there is only one table being returned
                elif PA_results[1] == 201:
                    results.append(
                        {
                            "isSuccess": True,
                            "results": PA_results[0]["data"],
                            "calc_unit_id": "",
                        }
                    )
                    display(
                        BasicWidgets.make_success_request_label(
                            text=f"PA Calculation Request Successfully Completed (Status Code: {PA_results[1]})"
                        )
                    )
                # if long running
                elif PA_results[1] == 202:
                    display(
                        BasicWidgets.make_header_label(
                            text=f"Retrieving Long Running Request: {first(PA_results).data.calculationid} (Status Code: {PA_results[1]})"
                        )
                    )
                    calculation_id = first(PA_results).data.calculationid
                    (
                        bcalc_completed,
                        status_response,
                    ) = FDSAnalyticsAPI.Requests.wait_for_long_running_response(
                        calculation_id=calculation_id,
                        pa_calculations_api=pa_calculations_api,
                    )
                    # check if long running request completed successfully
                    if bcalc_completed:
                        # Return Array or result JSONs
                        display(
                            BasicWidgets.make_success_request_label(
                                "PA Calculation Request Successfully Completed"
                            )
                        )
                        results = FDSAnalyticsAPI.Requests.return_long_running_response(
                            status_response=status_response,
                            calculation_id=calculation_id,
                            pa_calculations_api=pa_calculations_api,
                        )
                    else:
                        results.append(
                            {
                                "isSuccess": False,
                                "results": f"Request Failed:{str(status_response)}",
                                "calc_unit_id": "",
                            }
                        )
            else:
                self.output.clear_output()
                results.append(
                    {
                        "isSuccess": False,
                        "results": f"Request Failed: {PA_results}",
                        "calc_unit_id": "",
                    }
                )
                if unlinked_component_id:

                    FDSAnalyticsAPI.Requests.delete_unlinked_pa_template(
                        self.fds_config, id=unlinked_component_id
                    )

            for result in results:
                if result["isSuccess"]:
                    if result["calc_unit_id"]:
                        with self.output:
                            display(
                                BasicWidgets.make_success_request_label(
                                    f"Calculation Unit: {result['calc_unit_id']} Completed Successfully"
                                )
                            )
                    self.display_results(result["results"])
                    if return_results_as_json:
                        print(JSON.dumps(result["results"], indent=2))
                else:
                    display(
                        BasicWidgets.make_error_label(
                            alert_description=result["results"], alert_value=""
                        )
                    )

    def make_pa_delete_model_portfolio_request(self, model_portoflio):
        """Used to Display Steps to Delete Model Portfolio"""

        with self.output:
            display(
                BasicWidgets.make_request_label(
                    f"Deleting Model Portfolio: {model_portoflio}"
                )
            )
            (
                baccount_deleted,
                response_label,
                URL,
            ) = FDSAnalyticsAPI.Requests.delete_account(
                fds_username_hyphen_serial=self.__fds_username_hyphen_serial,
                fds_apikey=self.__fds_apikey,
                account_name_path=model_portoflio,
            )
            if baccount_deleted:
                display(BasicWidgets.make_header_label("DELETE REQUEST URL:"))
                print(URL)
                display(BasicWidgets.make_success_request_label(response_label))
            else:
                display(BasicWidgets.make_error_label(response_label, URL))

    def make_pa_delete_unliked_template_request(self, component_id):
        """Used To Display Steps to Delete Unlinked Template"""
        with self.output:
            display(
                BasicWidgets.make_request_label(
                    text="Deleting Unlinked PA Template Report"
                )
            )
            display(BasicWidgets.make_header_label(text="DELETE REQUEST URL:"))
            print(
                f"{self.AnalyticsAPIDefaultValues.unlinked_templates_url}/{component_id}"
            )
            
            (
                bdeleted,
                delete_results,
                delete_error
            ) = FDSAnalyticsAPI.Requests.delete_unlinked_pa_template(
                fds_config=self.fds_config, id=component_id
            )
            if bdeleted:
                display(
                    BasicWidgets.make_success_request_label(
                        f"Unlinked Template Deleted (Status Code 204)"
                    )
                )
            else:
                display(
                    BasicWidgets.make_error_label(
                        alert_description=delete_results,
                        alert_value=delete_error,
                    )
                )

    def get_date_values(self, date_object, allow_empty=False):
        """Used to Valid Date Parameters and return object"""

        error_message = "At a Minimum You must specifiy an End Date and Frequency."

        bstart_date, start_date = self.get_value_from_widget(
            date_object["start_date_textbox"]
        )

        bend_date, end_date = self.get_value_from_widget(
            date_object["end_date_textbox"]
        )
        bFrequency, frequency = self.get_value_from_widget(date_object["freq_dropdown"])

        if allow_empty and not bstart_date and not bend_date:
            return True, None

        if not bFrequency:
            return False, "You must select a frequency"

        if frequency.strip().upper() == "SINGLE":
            if (
                bend_date and bstart_date
            ) and end_date.strip().upper() != start_date.strip().upper():
                return False, "The Single frequency expects only an End Date"
            if bend_date:
                return True, FDSAnalyticsAPI.Paramaters.make_PA_date_parameters(
                    end_date=end_date, frequency=frequency
                )
        elif not (bend_date and bstart_date):
            return (
                False,
                "When choosing a frequency other than Single you must specifcy both a start and end date.",
            )
        return True, FDSAnalyticsAPI.Paramaters.make_PA_date_parameters(
            end_date=end_date, start_date=start_date, frequency=frequency
        )

    def get_group_values(self, group_box_selectmultiple):
        """
        Create a list containing a new PACalculationGroup for each option in selectmultiple
        Return true if at least one item is contained in the list
        """

        if group_box_selectmultiple.options:
            return True, [
                FDSAnalyticsAPI.Paramaters.make_PA_calculation_group(id=selection[1][3])
                for selection in list(group_box_selectmultiple.options)
            ]

        else:
            return False, []

    def get_PA_calculation_columns(self, columns_selectmultiple):
        """Get Selected PA Columns"""

        error_message = "You must select one or more columns"

        if columns_selectmultiple.options:
            return True, [
                FDSAnalyticsAPI.Paramaters.make_PA_calculation_column(
                    id=selection[1][2], statistics=selection[1][3]
                )
                for selection in list(columns_selectmultiple.options)
            ]
        else:
            return False, error_message

    def get_PA_data_source(self, pa_price_source_object):
        #  "portfolio_selectmultiple": portfolio_selectmultiple,
        #     "benchmark_selectmultiple": benchmark_selectmultiple,
        #     "check_box": check_box
        portfolio_pricing_source = self.get_pricing_source_list(
            pa_price_source_object["portfolio_selectmultiple"]
        )
        benchmark_pricing_source = self.get_pricing_source_list(
            pa_price_source_object["benchmark_selectmultiple"]
        )
        use_port_pricing = pa_price_source_object["check_box"].value
        return FDSAnalyticsAPI.Paramaters.make_PA_data_source(
            portfolio_pricing_sources=portfolio_pricing_source,
            benchmark_pricing_sources=benchmark_pricing_source,
            use_portfolio_pricing_source=use_port_pricing,
        )

    def get_pricing_source_list(self, selectmultiple):
        """Get Pricing Source IDs  from Select multiple"""

        if selectmultiple.options:
            return [
                FDSAnalyticsAPI.Paramaters.make_PA_princing_source(selection[1][3])
                for selection in list(selectmultiple.options)
            ]
        return []

    def get_PA_identifier_lst(self, selectmultiple):
        "Get PA Identifier List for Either Benchmark or Portfolios"
        if selectmultiple.options:
            return [
                FDSAnalyticsAPI.Paramaters.make_PA_identifier(
                    PA_Identifier=selection[1]
                )
                for selection in list(selectmultiple.options)
            ]
        return []

    def get_selectmultiple_values(self, selectmultiple):
        """
        Return Full List of Values stored inside of selectmultiple widget
        """

        if len(selectmultiple.options) > 0:
            return True, [selection[1][2] for selection in list(selectmultiple.options)]

        return False, []

    def get_value_from_widget(self, widget):
        """Return Value Paramater from widget (check if empty)"""

        if not widget.value:
            return False, ""
        else:
            return True, widget.value

    # Display Results Call On Truncate from Common Functions

    def display_results(self, result):
        # Display Calcuation Results inside of Widgets UI
        output_df = first(
            FDSAnalyticsAPI.FormatResults.return_calculation_results_as_dataframe(
                result
            )
        )
        (
            b_Truncated,
            df,
            rows_skipped_label,
        ) = FDSAnalyticsAPI.FormatResults.truncate_and_display_tables(
            output_df, max_rows_returned=self.max_rows_returned
        )
        with self.output:
            display(df)
            if b_Truncated:
                BasicWidgets.make_display_label(rows_skipped_label)

    class AnalyticsAPIDefaultValues:
        unlinked_templates_url = (
            "https://api.factset.com/analytics/engines/pa/v3/unlinked-templates"
        )
        calculations_url = (
            "https://api.factset.com/analytics/engines/pa/v3/calculations"
        )

        """Component Detail Box"""
        component_detail_label = (
            "Please select the level of granualarity for the Report."
        )

        # Possible Values: securities, groups, totals
        unlinked_template_key = "UNLINKED TEMPLATES"

        pricing_label = """Optionally, select one or more Pricing Sources to overwrite the default values stored within FactSet.  Selecting the checkbox will apply the same pricing sources to both the Portfolio and Benchmark.
         If you select the checkbox, but provide no Pricing Sources the default pricing sources from the document will be applied to both the portfolio and benchmark.
        
        """

        portfolio_benchmark_label = """Please enter the full path to at least one portfolio (FactSet Account, FactSet Composite, or Benchmark ID).
            When no benchmark is specified the API will use the default benchmark for the specified account or composite stored inside of Portfolio List Manager (@PLM).
            Listing more than one Portfolio and Benchmark will result in calculations being returned for each Portfolio and Benchmark Combination.
            """
        error_missing_portfolio = "You must specifiy at least one Portfolio."

        """Group Box"""
        group_label = "Optionally, specify one or more grouping to apply to the calculation. <br>*The grouping hierarchy is determined by the order they are listed."
        """Columns Box"""
        columns_label = """Please select one or more column(s) to be included in the dynamic calculation request.
            Optionally, set one or more statistic to be applied to each column individually.
            <br>*If no statistic is specified, the default statistic(s) stored inside FactSet will be used."""
        """Currency Box"""
        currency_label = (
            "Select what currency will be used to normalized the calculation results."
        )

        """Date Box"""
        date_description = "Start Date and End Date can be in either absolute (YYYYMMDD) or Relative Date Formats"
