from calendar import c
from Analytics_API_Common_Functions import FDSAnalyticsAPI
from IPython.display import display, clear_output
from Analytics_API_Common_Widgets import AnalyticsAPIWidgets
from Common_Widgets import (
    WidgetCombinations,
    BasicWidgets,
    WidgetContainers,
    WidgetLayouts,
    WidgetEvents,
)
from toolz import first
from functools import partial
import json as JSON

class ComponentManager(AnalyticsAPIWidgets):
    # UI Variables
    __general_info_object = None
    __component_detail_dropdwn = None
    __template_type_selectmultiple = None
    __portfolio_benchmark_object = None
    __portfolio_benchmark_object = None
    __group_box_selectmultiple = None
    __columns_selectmultiple = None
    __currecny_iso_dropdown = None
    __date_object = None
    __delete_component_checkbox = None
    __delete_model_portfolio_checkbox = None
    __all_in_one = False
    __has_model_portfolio = False
    __model_portfolio = None
    unlinked_template_tuple = None
    __return_results_as_json = False

    def __init__(
        self,
        fds_username_hyphen_serial,
        fds_apikey,
        all_in_one=True,
        model_portfolio=None,
        return_results_as_json = False
    ):
        super().__init__(
            fds_username_hyphen_serial=fds_username_hyphen_serial, fds_apikey=fds_apikey
        )
        if not model_portfolio is None:
            self.__model_portfolio = model_portfolio
            self.__has_model_portfolio = True

        if all_in_one:
            self.__all_in_one = True
        self.__return_results_as_json = return_results_as_json
        self.__renderUI()

    def __renderUI(self):
        create_submit_box = self.__make_create_submit_box()
        (general_info_box) = self.__make_general_info_box()
        (
            self.__component_detail_dropdwn,
            component_detail_box,
        ) = self.make_component_detail_box(
            self.default_values.component_detail_default_value
        )

        (template_type_box) = self.__make_template_type_box()

        if self.__all_in_one:
            (
                self.__portfolio_benchmark_object,
                portfolio_benchmark_box,
            ) = self.make_portfolio_benchmark_box(
                self.__model_portfolio
                if self.__has_model_portfolio
                else self.default_values.default_portfolio,
                self.default_values.default_benchmark,
            )
        (self.__group_box_selectmultiple, group_box) = self.make_groups_box(
            self.default_values.group_default_value
        )
        (self.__columns_selectmultiple, columns_box) = self.make_columns_box(
            self.default_values.columns_default_values
        )
        (self.__currecny_iso_dropdown, currency_iso_box) = self.make_currency_iso_box(
            self.default_values.currency_default_value
        )
        (self.__date_object, date_box) = self.make_dates_box(
            self.default_values.default_start_date,
            self.default_values.default_end_date,
            self.default_values.default_frequency,
        )
        (
            self.__pricing_source_object,
            pricing_source_box,
        ) = self.make_pricing_sources_box(self.default_values.pricing_source_default_values)

        if self.__all_in_one:
            childs = [
                general_info_box,
                component_detail_box,
                template_type_box,
                portfolio_benchmark_box,
                group_box,
                columns_box,
                currency_iso_box,
                date_box,
                pricing_source_box,
            ]
            tab_titles = [
                "General Info",
                "Component Detail",
                "Template Type",
                "Portfolio and Benchmark",
                "Groups",
                "Columns",
                "Currency",
                "Dates",
                "Pricing Sources",
            ]

        else:
            childs = [
                general_info_box,
                component_detail_box,
                template_type_box,
                group_box,
                columns_box,
                currency_iso_box,
                date_box,
                pricing_source_box,
            ]
            tab_titles = [
                "General Info",
                "Component Detail",
                "Template Type",
                "Groups",
                "Columns",
                "Currency",
                "Dates",
                "Pricing Sources",
            ]

        tab_box = WidgetContainers.make_vbox(
            children=[
                WidgetContainers.make_tab(
                    children=childs,
                    titles=tab_titles,
                ),
                create_submit_box,
            ],
            layout=WidgetLayouts.make_layout(
                width="max-content", height="600px", margin="5px"
            ),
        )
        display(tab_box, self.output)

    # Component Manager Containers (boxes)
    def __make_content_box(self):
        list_content = [
            "accounts",
            "benchmarks",
            "columns",
            "componentdetail",
            "currencyisocode",
            "dates",
            "groups",
        ]
        selectmultiple_layout = WidgetLayouts.make_layout(width="150px", height="125px")

        (
            mandatory_select_box,
            mandatory_selectmultiple,
        ) = WidgetCombinations.make_selectmultiple_with_label(
            "Mandatory", list_content, layout=selectmultiple_layout
        )
        (
            locked_select_box,
            locked_selectmultiple,
        ) = WidgetCombinations.make_selectmultiple_label(
            "Locked", list_content, layout=selectmultiple_layout
        )
        (
            optional_select_box,
            optional_selectmultiple,
        ) = WidgetCombinations.make_selectmultiple_label(
            "Optional", list_content, layout=selectmultiple_layout
        )
        locked_selectmultiple.observe(
            partial(
                WidgetEvents.filter_two_selelctMultiples,
                selectmulti_update=mandatory_selectmultiple,
                optional_selectmultiple=optional_selectmultiple,
                list_content=list_content,
            ),
            names="value",
        )
        mandatory_selectmultiple.observe(
            partial(
                WidgetEvents.filter_two_selelctMultiples,
                selectmulti_update=locked_selectmultiple,
                optional_selectmultiple=optional_selectmultiple,
                list_content=list_content,
            ),
            names="value",
        )

        return WidgetContainers.make_hbox(
            children=[mandatory_select_box, locked_select_box, optional_select_box]
        ), {
            "mandatory_select_box": mandatory_select_box,
            "locked_select_box": locked_select_box,
            "optional_select_box": optional_select_box,
        }

    def __make_create_submit_box(self):
        """'Create Execution Buttons"""
        run_lbl = "Run Calculation" if self.__all_in_one else "Make Component"

        run_button = BasicWidgets.make_button(run_lbl)
        return_json_button = BasicWidgets.make_button("Return JSON")
        children = [run_button, return_json_button]
        checkbox_box_children = []
        if self.__all_in_one:
            self.__delete_component_checkbox = BasicWidgets.make_checkbox(
                description="Delete Unlinked Template After Run"
            )
            checkbox_box_children.append(self.__delete_component_checkbox)
            if self.__has_model_portfolio:
                self.__delete_model_portfolio_checkbox = BasicWidgets.make_checkbox(
                    description="Delete Model Portfolio After Run"
                )
                checkbox_box_children.append(self.__delete_model_portfolio_checkbox)
            if len(checkbox_box_children) > 0:
                children.append(
                    WidgetContainers.make_hbox(
                        children=checkbox_box_children,
                        layout=WidgetLayouts.make_layout(max_width="475px"),
                    )
                )

        # Create Box
        create_submit_box = WidgetContainers.make_hbox(
            children=children,
            layout=WidgetLayouts.make_layout(height="40px", padding="0px 0px 0px 10px"),
        )
        # Add Events
        run_button.on_click(self.__run_button_click)
        return_json_button.on_click(self.__json_button_click)
        return create_submit_box

    def __make_general_info_box(self):

        description_label = BasicWidgets.make_description_label(
            self.default_values.general_info_label
        )
        directory_box, directory_textbox = WidgetCombinations.make_textbox_with_label(
            "Directory",
            value=self.default_values.general_info_default_directory,
            layout=WidgetLayouts.make_layout(width="400px", height="max-content"),
        )
        (
            description_box,
            description_textbox,
        ) = WidgetCombinations.make_textbox_with_label(
            label="Description",
            value=self.default_values.general_info_default_description,
            layout=WidgetLayouts.make_layout(width="400px", height="max-contnet"),
        )
        self.__general_info_object = {
            "directory_textbox": directory_textbox,
            "description_textbox": description_textbox,
        }

        return WidgetContainers.make_vbox(
            children=[description_label, directory_box, description_box],
            layout=WidgetLayouts.make_tab_box_layout(),
        )

    def __make_template_type_box(self):
        (
            b_template_types,
            response,response_err
        ) = FDSAnalyticsAPI.LookUp.get_default_unlinked_PA_template_types(
            fds_config=self.fds_config
        )

        template_types = FDSAnalyticsAPI.FormatResults.return_categorized_dictionary(
            response
        )
        if b_template_types:
            description_label = BasicWidgets.make_description_label(
                self.default_values.template_type_label
            )
            (
                template_type_box,
                self.__template_type_selectmultiple,
            ) = WidgetCombinations.make_category_item_box(
                components_dictionary=template_types,
                category_label="Template Type Categories",
                item_label="Template Types",
                selected_label="Selected Template",
                single_select=True,
            )
            self.__template_type_selectmultiple.options = (
                self.default_values.template_type_default_value
            )

            return WidgetContainers.make_vbox(
                children=[description_label, template_type_box],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
        else:
            return BasicWidgets.make_error_label(
                response, response_err
            )

    # Component Manager Workflows
    def __make_unlinked_pa_template_request(self):
        """First This function Confirms all Required Inputs (__return_and_display_unlinked_PA_tempate_json)
        Then makes the unlinked template request
        Confirms it completed successfully
        returns either True/False component ID
        """
        self.output.clear_output()
        with self.output:
            display(
                BasicWidgets.make_request_label(
                    text="Generating Unlinked Templated Report"
                )
            )
            (
                breqest_created,
                unlinked_PA_template_parameters,
            ) = self.__return_and_display_unlinked_PA_tempate_json()
            # Missing Required Argument was thrown just return False and no Component ID
            if breqest_created == False:
                return False, ""

            # Attempt to Create Unlinked PA Templated Report
            (
                btemplate_request,
                response_message,
                response_detail,
            ) = FDSAnalyticsAPI.Requests.make_unlinked_pa_template_request(
                fds_config=self.fds_config, data=unlinked_PA_template_parameters
            )
     
            unlinked_description = response_detail["data"]["description"]
            unlinked_id = response_detail["data"]["id"]
            unlinked_category = response_detail["data"]["category"]
            if not self.__all_in_one:
                self.unlinked_template_tuple = (
                    f"{unlinked_description} ({unlinked_id})",
                    (
                        unlinked_category,
                        unlinked_description,
                        unlinked_id,
                        self.AnalyticsAPIDefaultValues.unlinked_template_key,
                    ),
                )
            # If Unlinked Templated Component Failed to Generate Throw Error and Return False
            if btemplate_request == False:
                display(
                    BasicWidgets.make_error_label(
                        alert_description=response_message,
                        alert_value=response_detail,
                    )
                )
                return False, ""
            # Create Success Label
            display(
                BasicWidgets.make_success_request_label(
                    text=response_message
                )
            )
            display(BasicWidgets.make_header_label(text="REQUEST RESPONSE:"))
            print(JSON.dumps(response_detail,indent=2))
        #Return True and Component ID
        return True, unlinked_id

    def __make_pa_calculation_request(self, component_id):
        """Create an Abbridged PA Calc Request Only Passing in Portfolio and Benchmark Parameters"""

        with self.output:
            display(
                BasicWidgets.make_request_label("GENERATING PA CALCULATION REQUEST")
            )
            (
                bcalculation_parameters,
                calculation_parameters,
            ) = self.__get_PA_calculation_parameters(
                [component_id], self.__portfolio_benchmark_object
            )

            if bcalculation_parameters == False:
                display(calculation_parameters)
            else:
                display(BasicWidgets.make_header_label(text="POST REQUEST URL:"))
                #display(
                #    BasicWidgets.make_label(
                        #text=
                print(self.AnalyticsAPIDefaultValues.calculations_url
                    )
                #)
                display(
                    BasicWidgets.make_header_label(
                        text="POST REQUEST BODY:",
                    )
                )
                calculation_parameter_root = (
                    FDSAnalyticsAPI.Paramaters.make_PA_calculation_parameters_root(
                        calculation_parameters=calculation_parameters
                    )
                )
                print(JSON.dumps(calculation_parameter_root.to_dict(), indent=2))
                #display(calculation_parameter_root.to_dict())
                self.make_pa_calculation_request(
                    calculation_parameter_root, component_id,self.__return_results_as_json
                )

    def __return_and_display_unlinked_PA_tempate_json(self):
        """Used to Display JSON for creating unlinked PA Template interaction"""
        bSuccessful, data = self.__get_unlinked_PA_template_parameters()    
        if bSuccessful:
            with self.output:
                display(BasicWidgets.make_header_label(text="POST REQUEST URL:"))
                print(self.AnalyticsAPIDefaultValues.unlinked_templates_url)
                display(BasicWidgets.make_header_label(text="POST REQUEST BODY:"))
                print(JSON.dumps(FDSAnalyticsAPI.Paramaters.make_unlinked_PA_template_parameters_root(data=data, meta=None).to_dict(),indent=2))

        else:
            with self.output:
                display(data)
        return bSuccessful, data

    # Create Submit Button Events
    def __json_button_click(self, b):
        self.output.clear_output()
        self.__return_and_display_unlinked_PA_tempate_json()

    def __run_button_click(self, b):
        # Move to it's own function
        bunlinked_template, component_id = self.__make_unlinked_pa_template_request()
        if bunlinked_template:
            if self.__all_in_one:
                self.__make_pa_calculation_request(component_id)
            if self.__delete_component_checkbox:
                if self.__delete_component_checkbox.value:
                    if self.__all_in_one:
                        self.make_pa_delete_unliked_template_request(component_id)
                    else:
                        with self.output:
                            display(
                                BasicWidgets.make_error_label(
                                    "Skipped Deleting Unlinked Template:",
                                    "This Template Should be used in the next step",
                                )
                            )
            if self.__delete_model_portfolio_checkbox:
                if self.__delete_model_portfolio_checkbox.value:
                    self.make_pa_delete_model_portfolio_request(self.__model_portfolio)

    # Component Manager Return Results from Visual
    def __get_general_info_values(self):
        """Return Values Set Inside the General Info Tab of Widget
        Checkes for a valid FactSet Directory (Personal:) and description
        either returns True, path, description
        or False, Error Message, ""
        """

        fds_dir_path = self.__general_info_object["directory_textbox"].value
        fds_report_description = self.__general_info_object["description_textbox"].value

        if not fds_dir_path and not fds_report_description:
            return (
                False,
                "You must specify both a FactSet directory and Report Description",
                "",
            )
        elif not fds_report_description:
            return False, "You must specify a Report Description", ""
        elif not fds_dir_path:
            return (
                False,
                "You must specify a <strong>Directory Path</strong> (Personal:)",
                "",
            )
        else:
            if str(fds_dir_path).upper().strip() == "PERSONAL:":
                return True, fds_dir_path, fds_report_description
            else:
                return (
                    False,
                    "Currently all Unlinked PA Templates must be saved in <strong> Personal:</strong>",
                    "",
                )

    def __get_unlinked_PA_template_parameters(self):
        (
            bgeneral_info_values,
            fds_dir_path,
            fds_report_description,
        ) = self.__get_general_info_values()
        """ Returns All Parameters from UI to Create JSON To Request Unlinked Report Template"""
        btemplate_type, template_type_value = self.get_selectmultiple_values(
            selectmultiple=self.__template_type_selectmultiple
        )

        if bgeneral_info_values:
            if btemplate_type:
                template_parameters = (
                    FDSAnalyticsAPI.Paramaters.make_unlinked_PA_template_parameters(
                        directory=fds_dir_path,
                        template_type_id=first(template_type_value),
                        description=fds_report_description,
                    )
                )
                bDates, date_values = self.get_date_values(self.__date_object)
                if bDates:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=template_parameters,
                        name="dates",
                        object=date_values,
                    )
                else:
                    return False, BasicWidgets.make_error_label(
                        alert_description="Dates Type Error:", alert_value=date_values
                    )

                bGroups, group_values = self.get_group_values(
                    self.__group_box_selectmultiple
                )
                if not bGroups == False:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=template_parameters,
                        name="groups",
                        object=group_values,
                    )

                bCurrency, currencyISO = self.get_value_from_widget(
                    self.__currecny_iso_dropdown
                )

                if bCurrency:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=template_parameters,
                        name="currencyisocode",
                        object=currencyISO,
                    )
                bPricing, PA_Data_Sources = self.get_PA_data_source(
                    self.__pricing_source_object
                )
                if bPricing:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=template_parameters,
                        name="datasources",
                        object=PA_Data_Sources,
                    )
                bComponentDetail, component_detail = self.get_value_from_widget(
                    self.__component_detail_dropdwn
                )
                if bComponentDetail:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=template_parameters,
                        name="componentdetail",
                        object=component_detail,
                    )
                bColumns, PA_calculation_columns = self.get_PA_calculation_columns(
                    self.__columns_selectmultiple
                )
                if bColumns:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=template_parameters,
                        name="columns",
                        object=PA_calculation_columns,
                    )
                else:
                    return False, BasicWidgets.make_error_label(
                        alert_description="PA Calculation Column Error:",
                        alert_value=PA_calculation_columns,
                    )
            else:
                return False, BasicWidgets.make_error_label(
                    alert_description="Template Type Error:",
                    alert_value="You Must Select a Template Type ID.",
                )
        else:
            return False, BasicWidgets.make_error_label(
                alert_description="General Info Error:", alert_value=fds_dir_path
            )

        return True, template_parameters

    def __get_PA_calculation_parameters(
        self, component_ids, portfolio_benchmark_object
    ):
        """Gets Portfolio and Benchmark Info for the unlinked PA Template"""
        calculation_parameters = {}
        accounts_PA_identifier_list = self.get_PA_identifier_lst(
            self.__portfolio_benchmark_object["portfolio_selectmultiple"]
        )

        if accounts_PA_identifier_list:
            # for component_id in component_ids:
            calculation_parameter = (
                FDSAnalyticsAPI.Paramaters.make_PA_calculations_paramaters(
                    component_id=component_ids[0]
                )
            )
            FDSAnalyticsAPI.Paramaters.set_attribute(
                calculation_parameter, "accounts", accounts_PA_identifier_list
            )
            if portfolio_benchmark_object["benchmark_selectmultiple"]:
                FDSAnalyticsAPI.Paramaters.set_attribute(
                    calculation_parameter,
                    "benchmarks",
                    self.get_PA_identifier_lst(
                        portfolio_benchmark_object["benchmark_selectmultiple"]
                    ),
                )

            calculation_parameters["1"] = calculation_parameter
            return True, calculation_parameters
        else:
            return (
                False,
                BasicWidgets.make_error_label(
                    alert_description="Missing Account Error:",
                    alert_value=self.AnalyticsAPIDefaultValues.error_missing_portfolio,
                ),
            )

    class default_values:

        """General Info Box"""

        general_info_label = """Please specify the location (Directory) and name (Desciption) for the 
            Unlinked PA Template being created from this notebook."
            <br><br>*Currently all Unlinked PA Templates must be saved in the <strong>personal:</strong> directory."""
        general_info_default_directory = "Personal:"
        general_info_default_description = "Relative Weights Report"

        """Template Type Box"""
        template_type_label = """Please select one templateTypeId.
            The templateTypeId links the report template to a default report from inside 
            FactSet’s default template library. This parameter helps to organize your library of Unlinked
            Templates and assigns default arguments such as date, groupings, and columns to your Report Template. 
            <br>*All arguments specified inside of Report Template will
            overwrite the arguments defined by the templateTypeId."""
        template_type_default_value = [
            (
                "Weights (996E90B981AEE83F14029ED3D309FB3F03EC6E2ACC7FD42C22CBD5D279502CFD)",
                (
                    "Weights",
                    "Weights",
                    "996E90B981AEE83F14029ED3D309FB3F03EC6E2ACC7FD42C22CBD5D279502CFD",
                ),
            )
        ]
        """Group Box"""
        group_default_value = [
            (
                "Economic Sector - FactSet (76B9D1311717B495527B04570DC32A3D730AD6C1E91DE92CDBE65ADB5C757527)",
                (
                    "Factset",
                    "FactSet",
                    "Economic Sector - FactSet",
                    "76B9D1311717B495527B04570DC32A3D730AD6C1E91DE92CDBE65ADB5C757527",
                ),
            )
        ]
        """Columns Box"""
        columns_default_values = [
            (
                (
                    "Port. Ending Weight (44C486E34C7E1F5CE2B481D441BC1433C791070CE8E23596890102DA9785A1FD)",
                    (
                        "Portfolio/Position Data",
                        "Port. Ending Weight",
                        "44C486E34C7E1F5CE2B481D441BC1433C791070CE8E23596890102DA9785A1FD",
                        ["f25882fe0b724c1da563a99ce96085db"],
                    ),
                )
            ),
            (
                (
                    "Bench. Ending Weight (61F53FD426BF48DC0D7E756A7BCD1A7C96983A6E7CAAE13691F8056B978DB9C7)",
                    (
                        "Benchmark/Position Data",
                        "Bench. Ending Weight",
                        "61F53FD426BF48DC0D7E756A7BCD1A7C96983A6E7CAAE13691F8056B978DB9C7",
                        ["f25882fe0b724c1da563a99ce96085db"],
                    ),
                )
            ),
            (
                (
                    "Variation in Ending Weight (5DD1567659E91DDF4726A34F74D6A6F71627203B6A5A3AF07ED54AB328EEA158)",
                    (
                        "Variation/Variation in Weights",
                        "Variation in Ending Weight",
                        "5DD1567659E91DDF4726A34F74D6A6F71627203B6A5A3AF07ED54AB328EEA158",
                        ["f25882fe0b724c1da563a99ce96085db"],
                    ),
                )
            ),
        ]
        """Pricing Source Box """
        pricing_source_default_values = [
            (
                "FactSet - Equity (C5FC3FB610CB115857969E6C8BD7A3EA171D799C6D6B59F0FFB84D1CDFAB776F)",
                (
                    "Equity",
                    "FactSet",
                    "FactSet - Equity",
                    "C5FC3FB610CB115857969E6C8BD7A3EA171D799C6D6B59F0FFB84D1CDFAB776F",
                ),
            )
        ]

        """Currency Box"""
        currency_default_value = "USD"

        """Date Box"""
        default_start_date = "0D"
        default_end_date = "0D"
        default_frequency = "Single"
        """Component Detail Box"""
        component_detail_default_value = "securities"
        """Portfolio Benchmark Box"""
        default_portfolio = "lion:jmom"
        default_benchmark = "lion:IVV"
