from copy import deepcopy
from wsgiref.validate import InputWrapper
from Analytics_API_Common_Functions import FDSAnalyticsAPI
from IPython.display import display, clear_output
from Analytics_API_Common_Widgets import AnalyticsAPIWidgets
from Common_Widgets import (
    WidgetCombinations,
    BasicWidgets,
    WidgetContainers,
    WidgetLayouts,
)
from toolz import first
from functools import partial
import json as JSON


class PACalculationWidget(AnalyticsAPIWidgets):
    dict_documents = {}
    __component_id_selectmultiple = None
    __pricing_source_object = None
    __portfolio_benchmark_object = None
    __component_detail_dropdwn = None
    __date_object = None
    __group_box_selectmultiple = None
    __currecny_iso_dropdown = None
    __columns_selectmultiple = None
    __delete_component_checkbox = None
    __delete_model_portfolio_checkbox = None
    model_portfolio_deleted = False
    component_deleted = False
    __return_results_as_json = False

    # __run_button = {}
    # __return_json_button = {}
    __tab_titles = [
        "Component ID",
        "Portfolio and Benchmark",
        "Component Detail",
        "Dates",
        "Groups",
        "Currency",
        "Columns",
        "Pricing Sources",
    ]
    __default_unlinked_template_tuple = None
    __model_portfolio = None
    __has_model_portfolio = False

    def __init__(
        self,
        fds_username_hyphen_serial,
        fds_apikey,
        default_unlinked_template_tuple=None,
        model_portfolio=None,
        return_results_as_json = False
    ):
        super().__init__(
            fds_username_hyphen_serial=fds_username_hyphen_serial, fds_apikey=fds_apikey
        )
        self.__return_results_as_json = return_results_as_json
        self.__default_unlinked_template_tuple = default_unlinked_template_tuple
        if not model_portfolio is None:
            self.__model_portfolio = model_portfolio
            self.__has_model_portfolio = True
        self.__render_ui()

    def __render_ui(self):
        button_box = self.__make_create_submit_container()
        (
            self.__component_id_selectmultiple,
            component_id_box,
        ) = self.__make_component_ID_box()
        (
            self.__component_detail_dropdwn,
            component_detail_box,
        ) = self.make_component_detail_box(
            self.default_values.component_detail_default_value
        )
        (
            self.__portfolio_benchmark_object,
            portfolio_benchmark_box,
        ) = self.make_portfolio_benchmark_box(
            default_portfolio=self.__model_portfolio
            if self.__has_model_portfolio
            else self.default_values.default_portfolio,
            default_benchmark=self.default_values.default_benchmark,
        )
        (
            self.__pricing_source_object,
            pricing_source_box,
        ) = self.make_pricing_sources_box()

        (self.__group_box_selectmultiple, group_box) = self.make_groups_box(
            self.default_values.group_default_value
        )
        (self.__columns_selectmultiple, columns_box) = self.make_columns_box(
            self.default_values.columns_default_values
        )
        (self.__currecny_iso_dropdown, currency_iso_box) = self.make_currency_iso_box(
            self.default_values.currency_default_value
        )
        (self.__date_object, date_box) = self.make_dates_box(
            self.default_values.default_start_date,
            self.default_values.default_end_date,
            self.default_values.default_frequency,
        )

        tab_box = WidgetContainers.make_vbox(
            children=[
                WidgetContainers.make_tab(
                    children=[
                        component_id_box,
                        portfolio_benchmark_box,
                        component_detail_box,
                        date_box,
                        group_box,
                        currency_iso_box,
                        columns_box,
                        pricing_source_box
                    ],
                    titles=self.__tab_titles,
                ),
                button_box,
            ],
            layout=WidgetLayouts.make_layout(
                width="max-content", height="600px", margin="5px"
            ),
        )
        display(tab_box, self.output)

    # PA Calculation Containers (Boxes)
    def __make_component_ID_box(self):
        """Create a container that allows yout to return components for a specified PA Document"""

        if self.__default_unlinked_template_tuple:
            default_document = self.__default_unlinked_template_tuple[1][3]
            default_value = [self.__default_unlinked_template_tuple]
        else:
            default_document = self.default_values.default_PA_document
            default_value = self.default_values.component_id_default_value

        b_components, response, response_err = self.get_components(default_document)
        if b_components:
            descroption_label = BasicWidgets.make_description_label(
                self.default_values.component_ID_label
            )
            (
                category_item_box,
                selectmultiple,
                category_dropdown,
                item_dropdown,
                input_text,
                retrieve_button,
            ) = WidgetCombinations.make_category_item_box(
                components_dictionary=response,
                category_label="Report",
                item_label="Component/Tile",
                selected_label="Selected Component IDs",
                text_input_label="test",
                dict_info=self.dict_documents,
            )
            component_id_box = WidgetContainers.make_vbox(
                children=[descroption_label, category_item_box],
                layout=WidgetLayouts.make_tab_box_layout(),
            )
            input_text.value = default_document
            # retrieve_button.on_click(self.test)
            selectmultiple.options = default_value
            retrieve_button.on_click(
                partial(
                    self.__retrieve_button_clicked,
                    input_text=input_text,
                    category_dropdown=category_dropdown,
                    item_dropdown=item_dropdown,
                )
            )

            return selectmultiple, component_id_box
        else:
            return BasicWidgets.make_error_label(response, response_err)

    def __make_create_submit_container(self):
        run_button = BasicWidgets.make_button("Run Calculation")
        return_json_button = BasicWidgets.make_button("Return JSON")
        children = [run_button, return_json_button]
        checkbox_box_children = []

        if self.__default_unlinked_template_tuple:
            self.__delete_component_checkbox = BasicWidgets.make_checkbox(
                description="Delete Unlinked Template After Run"
            )
            checkbox_box_children.append(self.__delete_component_checkbox)
        if self.__has_model_portfolio:
            self.__delete_model_portfolio_checkbox = BasicWidgets.make_checkbox(
                description="Delete Model Portfolio After Run"
            )
            checkbox_box_children.append(self.__delete_model_portfolio_checkbox)
        if len(checkbox_box_children) > 0:
            children.append(
                WidgetContainers.make_hbox(
                    children=checkbox_box_children,
                    layout=WidgetLayouts.make_layout(max_width="475px"),
                )
            )

        button_box = WidgetContainers.make_hbox(
            children=children,
            layout=WidgetLayouts.make_layout(height="40px", margin="0px 0px 0px 10px"),
        )
        run_button.on_click(self.__run_button_click)
        return_json_button.on_click(self.__json_button_click)
        return button_box

    # PA Calcuation Workflows
    def __get_calculation_acct_bm_comp_id_info(self):
        """Returns True and calculation parameter when at least one account and component id has been selected"""
        accounts_PA_identifier_list = None
        benchmarks_PA_identifier_list = None
        bcomponent_ids, component_ids = self.get_selectmultiple_values(
            self.__component_id_selectmultiple
        )

        if bcomponent_ids:
            accounts_PA_identifier_list = self.get_PA_identifier_lst(
                self.__portfolio_benchmark_object["portfolio_selectmultiple"]
            )
            if accounts_PA_identifier_list:
                if self.__portfolio_benchmark_object["benchmark_selectmultiple"]:
                    benchmarks_PA_identifier_list = self.get_PA_identifier_lst(
                        self.__portfolio_benchmark_object["benchmark_selectmultiple"]
                    )
                return (
                    True,
                    component_ids,
                    accounts_PA_identifier_list,
                    benchmarks_PA_identifier_list,
                )
            else:
                return (
                    False,
                    BasicWidgets.make_error_label(
                        alert_description="Missing Account Error:",
                        alert_value=self.AnalyticsAPIDefaultValues.error_missing_portfolio,
                    ),
                    accounts_PA_identifier_list,
                    benchmarks_PA_identifier_list,
                )
        else:
            return (
                False,
                BasicWidgets.make_error_label(
                    alert_description="Component ID Error:",
                    alert_value=self.default_values.component_ID_error_message,
                ),
                accounts_PA_identifier_list,
                benchmarks_PA_identifier_list,
            )

    def __get_PA_calculation_parameters(self):
        """Returns All PA Calc Parameters from the UI"""
        calculation_parameters = {}
        (
            bcalculation_parameter,
            component_ids,
            accounts_PA_identifier_list,
            benchmark_PA_identifier_list,
        ) = self.__get_calculation_acct_bm_comp_id_info()

        if bcalculation_parameter:
            for component_id in component_ids:
                calculation_parameter = (
                    FDSAnalyticsAPI.Paramaters.make_PA_calculations_paramaters(
                        component_id=component_id
                    )
                )
                FDSAnalyticsAPI.Paramaters.set_attribute(
                    calculation_parameter, "accounts", accounts_PA_identifier_list
                )
                if benchmark_PA_identifier_list:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        calculation_parameter,
                        "benchmarks",
                        benchmark_PA_identifier_list,
                    )

                bDates, date_values = self.get_date_values(
                    self.__date_object, allow_empty=True
                )
                if bDates:
                    if date_values:
                        FDSAnalyticsAPI.Paramaters.set_attribute(
                            parameters=calculation_parameter,
                            name="dates",
                            object=date_values,
                        )
                else:
                    return False, BasicWidgets.make_error_label(
                        alert_description="Dates Type Error:", alert_value=date_values
                    )

                bGroups, group_values = self.get_group_values(
                    self.__group_box_selectmultiple
                )
                if not bGroups == False:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=calculation_parameter,
                        name="groups",
                        object=group_values,
                    )

                bCurrency, currencyISO = self.get_value_from_widget(
                    self.__currecny_iso_dropdown
                )

                if bCurrency:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=calculation_parameter,
                        name="currencyisocode",
                        object=currencyISO,
                    )

                bComponentDetail, component_detail = self.get_value_from_widget(
                    self.__component_detail_dropdwn
                )
                if bComponentDetail:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=calculation_parameter,
                        name="componentdetail",
                        object=component_detail,
                    )
                bColumns, PA_calculation_columns = self.get_PA_calculation_columns(
                    self.__columns_selectmultiple
                )
                if bColumns:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=calculation_parameter,
                        name="columns",
                        object=PA_calculation_columns,
                    )

                bPricing, PA_Data_Sources = self.get_PA_data_source(
                    self.__pricing_source_object
                )
                if bPricing:
                    FDSAnalyticsAPI.Paramaters.set_attribute(
                        parameters=calculation_parameter,
                        name="datasources",
                        object=PA_Data_Sources,
                    )

                calculation_parameters[
                    str(len(calculation_parameters) + 1)
                ] = calculation_parameter

            return True, calculation_parameters
        else:
            return bcalculation_parameter, component_ids

    def __return_and_display_PA_calculation_json(self):
        """Used for interaction when asking for just the JSON or the first step of requesting the calculation"""
        (
            bcalculation_parameters,
            calculation_parameters,
        ) = self.__get_PA_calculation_parameters()

        if bcalculation_parameters:
            with self.output:
                display(BasicWidgets.make_header_label(text="POST REQUEST URL:"))
                print(self.AnalyticsAPIDefaultValues.calculations_url)
                display(BasicWidgets.make_header_label(text="POST REQUEST Body:"))
                calculation_parameter_root = (
                    FDSAnalyticsAPI.Paramaters.make_PA_calculation_parameters_root(
                        calculation_parameters=calculation_parameters
                    )
                )
                print(JSON.dumps(calculation_parameter_root.to_dict(),indent=2))
                return True, calculation_parameter_root
        else:
            with self.output:
                display(calculation_parameters)
            return False, ""

    # Submit Button Events
    def __run_button_click(self, b):
        self.output.clear_output()
        with self.output:
            display(
                BasicWidgets.make_request_label("Generating PA Calculation Request")
            )
        (
            bcalculation_parameters,
            calculation_parameters_root,
        ) = self.__return_and_display_PA_calculation_json()
        if bcalculation_parameters:
            self.make_pa_calculation_request(
                calculation_parameter_root=calculation_parameters_root, return_results_as_json= self.__return_results_as_json
            )
        if self.__default_unlinked_template_tuple and self.__delete_component_checkbox:
            if self.__delete_component_checkbox.value:
                self.component_deleted = True
                self.make_pa_delete_unliked_template_request(
                    self.__default_unlinked_template_tuple[1][2]
                )

        if self.__has_model_portfolio and self.__delete_model_portfolio_checkbox:
            if self.__delete_model_portfolio_checkbox.value:
                self.model_portfolio_deleted = True
                self.make_pa_delete_model_portfolio_request(self.__model_portfolio)

    def __json_button_click(self, b):
        self.output.clear_output()
        self.__return_and_display_PA_calculation_json()

    def __retrieve_button_clicked(
        self, b, input_text, category_dropdown, item_dropdown
    ):
        PA_document_path = input_text.value.upper().strip()

        bComponents, response, response_err = self.get_components(PA_document_path=PA_document_path)
        if bComponents:
            category_dropdown.options = response
            item_dropdown.options = []
            with self.output:
                display(BasicWidgets.make_success_request_label("Components Successfully Updated"))  
        
        else:
            with self.output:
                display(BasicWidgets.make_error_label(
                        alert_description=response,
                        alert_value=response_err,
                    ))


    def get_components(self, PA_document_path):
        """Get Unlinked PA Templates or Component ID's for a specified document"""
        if PA_document_path == self.AnalyticsAPIDefaultValues.unlinked_template_key:
            (
                b_components,
                response,
                response_err,
            ) = FDSAnalyticsAPI.LookUp.get_unlinked_pa_templates(self.fds_config)
            if b_components:
                components = (
                    FDSAnalyticsAPI.FormatResults.return_categorized_dictionary(
                        pa_dictionary=response,
                        key_value=self.AnalyticsAPIDefaultValues.unlinked_template_key,
                        display_value="description",
                    )
                )
                self.dict_documents[PA_document_path] = components
            else:
                return False, response, response_err

        else:
            if PA_document_path not in self.dict_documents:
                (
                    b_components,
                    response, response_err
                ) = FDSAnalyticsAPI.LookUp.get_components(
                    fds_config=self.fds_config, pa3_document=PA_document_path
                )
                if b_components:
                    components = (
                            FDSAnalyticsAPI.FormatResults.return_categorized_dictionary(
                            response,
                            key_value=PA_document_path,
                            )
                        )
                    self.dict_documents[PA_document_path] = components
                else:
                    return False, response, response_err
            
            else:
                components = self.dict_documents[PA_document_path]

        return True, components, ""

    class default_values:
        """Component ID Box"""

        default_PA_document = "PA3_DOCUMENTS:DEFAULT"
        component_id_default_value = [
            (
                "Weights (801B800245E468A52AEBEC4BE31CFF5AF82F371DAEF5F158AC2E98C2FA324B46)",
                (
                    "Weights / Exposures",
                    "Weights",
                    "801B800245E468A52AEBEC4BE31CFF5AF82F371DAEF5F158AC2E98C2FA324B46",
                    "PA3_DOCUMENTS:DEFAULT",
                ),
            )
        ]

        component_ID_label = f"""Select one or more Component IDs to be included in your PA Engine Calculation Request.  
                                To return the component IDs for a specific document enter the full path to the document in the textbox below. 
                                 To return Unlinked PA Template component IDs enter {AnalyticsAPIWidgets.AnalyticsAPIDefaultValues.unlinked_template_key} in the same textbox."""

        component_ID_error_message = "You must specify at least one component ID."
        """Group Box"""

        group_default_value = []  # [
        #     (
        #         "Economic Sector - FactSet (76B9D1311717B495527B04570DC32A3D730AD6C1E91DE92CDBE65ADB5C757527)",
        #         (
        #             "FactSet",
        #             "Economic Sector - FactSet",
        #             "76B9D1311717B495527B04570DC32A3D730AD6C1E91DE92CDBE65ADB5C757527",
        #         ),
        #     )
        # ]
        """Columns Box"""
        columns_default_values = []  # [
        #     (
        #         (
        #             "Port. Ending Weight (44C486E34C7E1F5CE2B481D441BC1433C791070CE8E23596890102DA9785A1FD)",
        #             (
        #                 "Portfolio/Position Data",
        #                 "Port. Ending Weight",
        #                 "44C486E34C7E1F5CE2B481D441BC1433C791070CE8E23596890102DA9785A1FD",
        #                 ["f25882fe0b724c1da563a99ce96085db"],
        #             ),
        #         )
        #     ),
        #     (
        #         (
        #             "Bench. Ending Weight (61F53FD426BF48DC0D7E756A7BCD1A7C96983A6E7CAAE13691F8056B978DB9C7)",
        #             (
        #                 "Benchmark/Position Data",
        #                 "Bench. Ending Weight",
        #                 "61F53FD426BF48DC0D7E756A7BCD1A7C96983A6E7CAAE13691F8056B978DB9C7",
        #                 ["f25882fe0b724c1da563a99ce96085db"],
        #             ),
        #         )
        #     ),
        #     (
        #         (
        #             "Variation in Ending Weight (5DD1567659E91DDF4726A34F74D6A6F71627203B6A5A3AF07ED54AB328EEA158)",
        #             (
        #                 "Variation/Variation in Weights",
        #                 "Variation in Ending Weight",
        #                 "5DD1567659E91DDF4726A34F74D6A6F71627203B6A5A3AF07ED54AB328EEA158",
        #                 ["f25882fe0b724c1da563a99ce96085db"],
        #             ),
        #         )
        #     ),
        # ]
        """Currency Box"""
        currency_default_value = ""  # "USD"
        """Date Box"""
        default_start_date = ""  # "0D"
        default_end_date = ""  # "0D"
        default_frequency = ""  # "Single"
        """Component Detail Box"""
        component_detail_default_value = ""  # "securities"
        """Portfolio Benchmark Box"""
        default_portfolio = "lion:jmom"
        default_benchmark = "lion:IVV"
